#!/usr/bin/env python

# Copyright 2018 Earth Sciences Department, BSC-CNS
#
# This file is part of HERMESv3_GR.
#
# HERMESv3_GR is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HERMESv3_GR is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HERMESv3_GR. If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import numpy as np
import timeit
import ESMF

import hermesv3_gr.config.settings as settings
from regrid import Regrid

# from mpi4py import MPI
# icomm = MPI.COMM_WORLD
# comm = icomm.Split(color=0, key=0)
# nprocs = comm.Get_size()
# rank = comm.Get_rank()


class ConservativeRegrid(Regrid):
    """
    ConservativeRegrid class that contains all the needed information to do the conservative regrid.

    :param input_file: Path to the input emission file.
    :type input_file: str

    :param pollutants: Pollutants available of the input file.
    :type pollutants: list

    :param weight_matrix_file: Path to the weight matrix that will be created if it is not created yet.
    :type weight_matrix_file: str

    :param grid: ...
    :type grid: Grid

    :param masking: ...
    :type masking: Masking
    """
    def __init__(self, pollutant_dicts, weight_matrix_file, grid, masking=None):
        st_time = timeit.default_timer()
        settings.write_log('\t\tInitializing Conservative regrid.', level=2)

        super(ConservativeRegrid, self).__init__(pollutant_dicts, weight_matrix_file, grid, masking=masking)

        settings.write_time('ConservativeRegrid', 'Init', timeit.default_timer() - st_time, level=2)

        return None

    def create_weight_matrix(self):
        """
        Calls to ESMF_RegridWeightGen to generate the weight matrix.
        """

        st_time = timeit.default_timer()

        src_grid = self.grid.create_esmf_grid_from_file(self.pollutant_dicts[0]['path'])
        src_field = ESMF.Field(src_grid, name='my input field')
        src_field.read(filename=self.pollutant_dicts[0]['path'], variable=self.pollutant_dicts[0]['name'], timeslice=0)

        dst_grid = self.grid.esmf_grid
        dst_field = ESMF.Field(dst_grid, name='my outut field')
        regrid = ESMF.Regrid(src_field, dst_field, filename=self.weight_matrix_file, regrid_method=ESMF.RegridMethod.CONSERVE, )#src_mask_values=self.masking.regrid_mask)
        # regrid = ESMF.Regrid(src_field, dst_field, filename=self.weight_matrix_file, regrid_method=ESMF.RegridMethod.BILINEAR, )#src_mask_values=self.masking.regrid_mask)

        settings.write_time('ConservativeRegrid', 'create_weight_matrix', timeit.default_timer() - st_time, level=1)

    def start_regridding(self, gfas=False, vertical=None):
        from hermesv3_gr.tools.netcdf_tools import extract_vars

        st_time = timeit.default_timer()

        weights = self.read_weight_matrix()

        dst_field_list = []
        num =1
        for pollutant_single_dict in self.pollutant_dicts:
            settings.write_log('\t\tPollutant {0} ({1}/{2})'.format(
                pollutant_single_dict['name'], num, len(self.pollutant_dicts)), level=3)
            num += 1

            [values] = extract_vars(pollutant_single_dict['path'], [pollutant_single_dict['name']])
            values = values['data']
            if gfas:
                values = vertical.do_vertical_interpolation_allocation(values, vertical.altitude)
            # Do masking
            if self.masking.regrid_mask is not None:
                values = np.multiply(values, self.masking.regrid_mask, dtype=settings.precision)
            # Do scalling
            if self.masking.scale_mask is not None:
                values = np.multiply(values, self.masking.scale_mask, dtype=settings.precision)
            if gfas:
                values = values.reshape((values.shape[-3], values.shape[-2] * values.shape[-1],))
            else:
                values = values.reshape((1, values.shape[-2] * values.shape[-1],))

            unique, counts = np.unique(weights['row'], return_counts=True)
            new_dst_indices = np.cumsum(counts)

            # Expand src values
            src_aux = np.take(values, weights['col'], axis=1)

            # Apply weights
            dst_field_aux = np.multiply(src_aux, weights['S'], dtype=settings.precision)

            # Reduce dst values
            dst_field = self.reduce_dst_field(dst_field_aux, new_dst_indices, self.grid.shape[-1] * self.grid.shape[-2])

            if gfas:
                dst_field = vertical.do_vertical_interpolation(dst_field)
                dst_field = dst_field.reshape((self.grid.shape[-3], self.grid.shape[-2], self.grid.shape[-1],))
            else:
                dst_field = dst_field.reshape((self.grid.shape[-2], self.grid.shape[-1],))

            dst_field_list.append({'data': dst_field, 'name': pollutant_single_dict['name']})

        settings.write_time('ConservativeRegrid', 'start_regridding', timeit.default_timer() - st_time, level=3)
        return dst_field_list

    # def start_carles_regridding_1(self):
    #     from hermesv3_gr.tools.netcdf_tools import extract_vars
    #
    #     weights = self.read_weight_matrix()
    #     print 'Weights Rank {3} col: {0} row: {1} S: {2}'.format(weights['col'].shape, weights['row'].shape,
    #                                                              weights['S'].shape, settings.rank)
    #
    #     dst_field_list = []
    #     for pollutant_single_dict in self.pollutant_dicts:
    #         if settings.log_level_2:
    #             print '\t\t {0}'.format(pollutant_single_dict['name'])
    #         # dst_field = np.zeros(((self.grid.x_upper_bound - self.grid.x_lower_bound)*(self.grid.y_upper_bound - self.grid.y_lower_bound),))
    #         dst_field = np.zeros((weights['max'],))
    #
    #         [values] = extract_vars(pollutant_single_dict['path'], [pollutant_single_dict['name']])
    #         values = values['data'].reshape((values['data'].shape[-1]*values['data'].shape[-2],))
    #         # print 'VALUES SUM {0}'.format(values.sum())
    #         # print  dst_field.shape, weights['row'].max()
    #         # print values.shape
    #         # exit()
    #         # print 'Weights Rank {0} dst_field: {1} '.format(rank, dst_field.shape)
    #         # if rank == 1:
    #         if True:
    #             for i in xrange(len(weights['S'])):
    #                 try:
    #                     dst_field[weights['row'][i] - 1] += weights['S'][i] * values[weights['col'][i] - 1]
    #                     # print '{0}/{1} {2}%'.format(i+1, len(weights['S']), round((i+1)*100/len(weights['S'])),2)
    #                     # print values[weights['col'][i] - 1], weights['col'][i] - 1
    #                     # print dst_field[weights['row'][i] - 1], weights['S'][i], values[weights['col'][i] - 1]
    #                 except:
    #                     print values[weights['col'][i] - 1]
    #                     print weights['S'][i]
    #                     print dst_field[weights['row'][i] - 1]
    #         dst_field = dst_field.reshape(len(self.grid.center_latitudes),len(self.grid.center_longitudes))
    #         dst_field = dst_field[self.grid.x_lower_bound:self.grid.x_upper_bound, self.grid.y_lower_bound:self.grid.y_upper_bound]
    #
    #         print 'RESHAPE: {0}'.format((self.grid.x_upper_bound - self.grid.x_lower_bound, self.grid.y_upper_bound - self.grid.y_lower_bound))
    #         dst_field_list.append({'data': dst_field, 'name': pollutant_single_dict['name']})
    #
    #         # dst_field_list.append({'data': dst_field.reshape((self.grid.x_upper_bound - self.grid.x_lower_bound, self.grid.y_upper_bound - self.grid.y_lower_bound)), 'name': pollutant_single_dict['name']})
    #         # dst_field_list.append({'data': dst_field.reshape((self.grid.y_upper_bound - self.grid.y_lower_bound,self.grid.x_upper_bound - self.grid.x_lower_bound,)), 'name': pollutant_single_dict['name']})
    #     return dst_field_list

    def read_weight_matrix(self):
        from netCDF4 import Dataset
        dict_aux = {}
        nc = Dataset(self.weight_matrix_file, mode='r')

        dict_aux['col'] = nc.variables['col'][:]
        dict_aux['row'] = nc.variables['row'][:]
        dict_aux['S'] = nc.variables['S'][:]
        nc.close()
        dict_aux['max'] = dict_aux['row'].max()

        dict_aux['col'] -= 1
        dict_aux['row'] -= 1

        if settings.size != 1:
            inc = dict_aux['row'][:-1] - dict_aux['row'][1:]
            index = np.where(inc > inc.max() * 0.5)[0]
            index = np.concatenate([[0], index])

            try:
                if settings.rank != 0:
                    dict_aux['col'] = dict_aux['col'][index[settings.rank] +1: index[settings.rank + 1] + 1]
                    dict_aux['row'] = dict_aux['row'][index[settings.rank] +1: index[settings.rank + 1] + 1]
                    dict_aux['S'] = dict_aux['S'][index[settings.rank] +1: index[settings.rank + 1] + 1]
                else:
                    dict_aux['col'] = dict_aux['col'][: index[settings.rank + 1] + 1]
                    dict_aux['row'] = dict_aux['row'][: index[settings.rank + 1] + 1]
                    dict_aux['S'] = dict_aux['S'][: index[settings.rank + 1] + 1]
            except IndexError:
                dict_aux['col'] = dict_aux['col'][index[settings.rank] + 1:]
                dict_aux['row'] = dict_aux['row'][index[settings.rank] + 1:]
                dict_aux['S'] = dict_aux['S'][index[settings.rank] + 1:]

        return dict_aux

    def wait_to_weightmatrix(self):
        import time

        find = False

        while not find:
            if os.path.exists(self.weight_matrix_file):
                pre_size = 0
                post_size = 1
                print "I'm {0}".format(settings.rank), 'Writing Weight Matrix {0}'.format(self.weight_matrix_file)
                # find = True
                while pre_size != post_size:
                    print "I'm {0}".format(settings.rank), pre_size, post_size
                    pre_size = post_size
                    post_size = os.path.getsize(self.weight_matrix_file)
                    time.sleep(1)
                find = True
                print "I'm {0}".format(settings.rank), 'FINISHED'
            else:
                time.sleep(5)
                print "I'm {0}".format(settings.rank), 'Waiting Weight Matrix'

    def re_order_weight_matrix(self):
        """
        Takes the ESMF Weight Matrix and re-order it in our way to re-write it.

        It will re-order and re-name the variables:

        dst_indices = ordered (row)
        src_indices = ordered (col)
        weights = ordered (S)

        dst_indices_count:
            dst_indices =       [0, 0, 0, 1, 1, 1, 2, 2, 2, 2]
            dst_indices_count = [      3,       6,          10]

        """
        from netCDF4 import Dataset
        import numpy as np

        if settings.log_level_3:
            st_time = timeit.default_timer()

        # Read ESMF Weight matrix NetCDF
        nc_weights = Dataset(self.weight_matrix_file, mode='r')

        row = nc_weights.variables['row'][:]
        col = nc_weights.variables['col'][:]
        s = nc_weights.variables['S'][:]

        nc_weights.close()

        # Re-order and re-name values
        indices = np.argsort(row)
        dst_indices = np.array(row)[indices] - 1
        weights = np.array(s, dtype=settings.precision)[indices]
        src_indices = np.array(col)[indices] - 1

        # Short indices
        # dst_indices =  [0, 0, 0, 1, 1, 1, 2, 2, 2, 2]
        # new_dst_indices = [3, 6, 10]
        unique, counts = np.unique(dst_indices, return_counts=True)
        new_dst_indices = np.cumsum(counts)

        # Create new weight matrix NetCDF
        nc_weights = Dataset(self.weight_matrix_file, mode='w')

        nc_weights.createDimension('dim', len(indices))
        nc_weights.createDimension('dim2', new_dst_indices.shape[0])

        nc_dst_indices = nc_weights.createVariable('dst_indices', 'i', ('dim',), zlib=True)
        nc_dst_indices[:] = dst_indices

        nc_dst_indices_count = nc_weights.createVariable('dst_indices_count', 'i', ('dim2',), zlib=True)
        nc_dst_indices_count[:] = new_dst_indices

        nc_src_indices = nc_weights.createVariable('src_indices', 'i', ('dim',), zlib=True)
        nc_src_indices[:] = src_indices

        weights_var = nc_weights.createVariable('weights', 'd', ('dim',), zlib=True)
        weights_var[:] = weights

        nc_weights.close()

        if settings.log_level_3:
            print 'TIME -> ConservativeRegrid.re_order_weight_matrix: {0} s'.format(round(timeit.default_timer() - st_time, 2))

    def apply_weights_test(self, values):
        """
        Calculates the regridded values using the ESMF algorithm for a 3D array specifically for a conservative regrid.

        :param values: Input values to regrid.
        :type values: numpy.ndarray

        :return: Values already regridded.
        :rtype: numpy.ndarray
        """
        from netCDF4 import Dataset
        if settings.log_level_3:
            st_time = timeit.default_timer()
        dst_field = super(ConservativeRegrid, self).apply_weights(values)

        nc_weights = Dataset(self.weight_matrix_file, mode='r')

        n_b = nc_weights.dimensions['n_b'].size
        frac_b = nc_weights.variables['frac_b'][:]

        nc_weights.close()

        # ! Adjust destination field by fraction
        # do i=1, n_b
        #   if (frac_b(i) .ne. 0.0) then
        #      dst_field(i)=dst_field(i)/frac_b(i)
        #   endif
        # enddo
        for i in xrange(n_b):
            if frac_b[i] != 0:
                dst_field[:, i] = dst_field[:, i] / frac_b[i]
        if settings.log_level_3:
            print 'TIME -> ConservativeRegrid.apply_weights: {0} s'.format(round(timeit.default_timer() - st_time, 2))
        return dst_field

    def apply_weights(self, values):
        """
        Calculates the regridded values using the ESMF algorithm for a 3D array specifically for a conservative regrid.

        :param values: Input values to regrid.
        :type values: numpy.array

        :return: Values already regridded.
        :rtype: numpy.array
        """

        dst_field = super(ConservativeRegrid, self).apply_weights(values)

        return dst_field


if __name__ == '__main__':
    pass

